ALL_ASSET_FLOW_DATA_POINTS = [
    "TNA0M",
    "TNA0R",
    "ENF0M",
    "ENF0R",
    "OGR0R",
    "MAP0R",
    "TNA1M",
    "TNA1R",
    "ENF1M",
    "ENF1R",
    "OGR1R",
    "MAP1R",
    "ENFM1",
    "ENFQD",
    "ENFYD",
    "ENFTM",
    "ENF24",
    "ENF36",
]

FORMAT_DATE = "%Y-%m-%d"

# Some function require these data point settings in order to successfully execute without errors.
REQUIRED_DATA_POINT_SETTINGS = [
    "datapointId",
    "datapointName",
    "isTsdp",
    "alias",
    "displayName",
]

VISIBLE_DATA_POINT_SETTINGS = [
    "currency",
    "preEuroConversion",
    "sourceId",
    "frequency",
    "startDate",
    "endDate",
    "floatStart",
    "floatEnd",
    "startDelay",
    "endDelay",
    "diffStart",
    "diffEnd",
    "compounding",
    "calculationId",
    "annualized",
    "annualDays",
    "benchmark",
    "riskfree",
    "windowType",
    "windowSize",
    "stepSize",
    "requireContinueData",
    "fit",
    "scalType",
    "scalValue",
    "scalPercentValue",
    "timehorizon",
]

HIDDEN_DATA_POINT_SETTINGS = [
    "type",
    "isFilterable",
    "isEpdp",
    "isCustomCalc",
    "canBeAddedToDataset",
    "mstarIpType",
    "nonstandardDisplayType",
    "hasOptions",
    "requireValueSearch",
    "scale",
    "reverseSign",
    "decimalPlaces",
    "displayOptions",
    "align",
    "width",
    "cellType",
    "sortLocal",
    "displayType",
    "percentileRank",
    "hide",
    "custom",
    "fixFormat",
    "customType",
    "generalFrequency",
    "calcCurType",
    "calcSdType",
    "calcUse5Days",
    "calcS1",
    "calcS2",
    "calcCfgDataSource",
    "calcBestReturnNum",
    "calcWorstReturnNum",
    "calcRollingTimePeriod",
    "calcReinvestDateType",
    "calcReinvestSource",
    "calcAdditionalSource",
    "calcIsApplyIsraelsenModification",
    "calcDaysOfAnnuYear",
    "calcMnav",
    "showType",
    "transType",
    "searchAttributes",
    "additionalReqId",
    "isPortfolioCalc",
    "asOfDate",
    "universe",
]
