investments_list_default_columns = [
    "secid",
    "masterportfolioid",
    "tradingsymbol",
    "name",
    "securitytype",
    "exchangeid",
    "category",
]

# these datapointIds correspond to the above columns
investments_list_data_points = [
    {"datapointId": "OS00I", "isTsdp": False},
    {"datapointId": "DC09A", "isTsdp": False},
    {"datapointId": "OS385", "isTsdp": False},
    {"datapointId": "OS01W", "isTsdp": False},
    {"datapointId": "OS010", "isTsdp": False},
    {"datapointId": "OS01Z", "isTsdp": False},
    {"datapointId": "OS245", "isTsdp": False},
]
